import os
from pathlib import Path
from src.pycov3.Directory import Cov3Dir

window_params = {
    "window_size": 500,
    "window_step": 10,
    "edge_length": 10,
}
coverage_params = {
    "mapq_cutoff": None,
    "mapl_cutoff": None,
    "max_mismatch_ratio": None,
}
window_params = {k: v for k, v in window_params.items() if v is not None}
coverage_params = {k: v for k, v in coverage_params.items() if v is not None}

fns = ["max_bin.001", "max_bin.002"]


def test_cov3_dir_init():
    # Define a temporary directory path for testing
    temp_dir = Path("cov3_dir")
    os.makedirs(temp_dir)

    cov3_dir = Cov3Dir(temp_dir, False, fns, window_params, coverage_params)

    assert len(cov3_dir.files) == 2


def test_cov3_dir_get_bin():
    # Define a temporary directory path for testing
    temp_dir = Path("cov3_dir")
    os.makedirs(temp_dir)

    cov3_dir = Cov3Dir(temp_dir, False, fns, window_params, coverage_params)

    assert cov3_dir.get_bin("001").bin_name == "001"



import shutil
import sys
import pytest
from pathlib import Path


# each test runs on cwd to its temp dir
@pytest.fixture(autouse=True)
def go_to_tmpdir(request):
    base_fp = Path([fp for fp in sys.path if fp.endswith("pycov3")][0])

    # Get the fixture dynamically by its name.
    tmpdir = request.getfixturevalue("tmpdir")
    # ensure local test created packages can be imported
    sys.path.insert(0, str(tmpdir))
    # Chdir only for the duration of the test.
    with tmpdir.as_cwd():
        shutil.copytree(base_fp / "tests" / "data" / "sim", Path(tmpdir) / "sim")
        yield



import os
from pathlib import Path
from src.pycov3.Directory import SamDir
from src.pycov3.File import SamFile, FastaFile
from src.pycov3.App import Cov3Generator
from src.pycov3.Sequence import Contig
from tests.unit.utils import (
    create_sample_fasta_file,
    create_sample_sam_file,
)


def test_cov3_generator_generate_cov3():
    # Test generate_cov3() method of Cov3Generator
    sam_file_path = Path("sams/Akk_001.sam")
    os.makedirs("sams", exist_ok=True)
    create_sample_sam_file(sam_file_path)
    sam_dir = SamDir(Path("sams"), True)

    fasta_file_path = Path("max_bin.001.fa")
    create_sample_fasta_file(fasta_file_path)
    fasta_file = FastaFile(fasta_file_path)

    cov3_generator = Cov3Generator(
        {s.fp.stem: s.parse() for s in sam_dir.files},
        fasta_file.parse(),
        "Akk",
        "001",
        {
            "window_size": 500,
            "window_step": 10,
            "edge_length": sam_dir.calculate_edge_length(),
        },
        1,
        30,
        0.03,
    )

    assert (
        list(cov3_generator.generate_cov3()) == []
    )  # Test data is too small for window_size min


def test_cov3_file_update_coverages():
    # Test cov3 file update_coverages utility
    sam_file_path = Path("Akk_001.sam")
    create_sample_sam_file(sam_file_path)
    sam_file = SamFile(sam_file_path)
    sam_lines = list(sam_file.parse())

    cov3_generator = Cov3Generator(
        [sam_file.parse()],
        None,
        "Akk",
        "001",
        {"window_size": 500, "window_step": 10, "edge_length": 5},
        1,
        30,
        0.03,
    )

    coverages = {}

    for line in sam_lines:
        coverages = cov3_generator._Cov3Generator__update_coverages(
            coverages, line, 2, 2
        )

    assert coverages == {
        -1: 2,
        14: 2,
        0: 2,
        1: 2,
        2: 2,
        3: 2,
        4: 4,
        5: 4,
        6: 4,
        7: 4,
        8: 4,
        9: 4,
        10: 4,
        11: 4,
        12: 4,
        13: 4,
        19: 0,
        15: 2,
        16: 2,
        17: 2,
        18: 2,
    }


def test_cov3_file_log_cov_info():
    # Test cov3 file update_coverages utility
    fasta_file_path = Path("max_bin.001.fa")
    create_sample_fasta_file(fasta_file_path)
    fasta_file = FastaFile(fasta_file_path)
    fasta_records = list(fasta_file.parse())

    cov3_generator = Cov3Generator(
        [],
        fasta_file.parse(),
        "Akk",
        "001",
        {"window_size": 500, "window_step": 10, "edge_length": 5},
        1,
        30,
        0.03,
    )

    contig = Contig(
        fasta_records[0][0], fasta_records[0][1] * 10, "Akk0", "001", 5, 500, 10
    )
    coverages = {}
    for i in range(-1, 599):
        coverages[i] = 2
    coverages[599] = 0

    info = cov3_generator._Cov3Generator__log_cov_info(contig, coverages, 5, 500, 10)
    assert info == [
        {"log_cov": -2.3219, "GC_content": 0.496},
        {"log_cov": -2.3219, "GC_content": 0.496},
        {"log_cov": -2.3219, "GC_content": 0.5},
        {"log_cov": -2.3219, "GC_content": 0.504},
        {"log_cov": -2.3219, "GC_content": 0.504},
        {"log_cov": -2.3219, "GC_content": 0.502},
        {"log_cov": -2.3219, "GC_content": 0.498},
        {"log_cov": -2.3219, "GC_content": 0.496},
        {"log_cov": -2.3219, "GC_content": 0.496},
        {"log_cov": -2.3219, "GC_content": 0.5},
        {"log_cov": -2.3219, "GC_content": 0.504},
        {"log_cov": -2.3219, "GC_content": 0.504},
        {"log_cov": -2.3219, "GC_content": 0.502},
        {"log_cov": -2.3219, "GC_content": 0.498},
        {"log_cov": -2.3219, "GC_content": 0.496},
        {"log_cov": -2.3219, "GC_content": 0.498},
        {"log_cov": -2.3219, "GC_content": 0.502},
    ]


def test_cov3_file_calculate_mapl():
    # Test cov3 file mapping length calculation
    assert Cov3Generator.calculate_mapl("250M") == 250
    assert Cov3Generator.calculate_mapl("100M50I100M") == 150
    assert Cov3Generator.calculate_mapl("200D50I") == 150
    assert Cov3Generator.calculate_mapl("250H") == 0
    assert Cov3Generator.calculate_mapl("*") == -1



import logging
from abc import ABC, abstractmethod
from itertools import groupby
from pathlib import Path
from typing import Dict, Iterator, List, Tuple, Union

from .App import Cov3Generator


class File(ABC):
    def __init__(self, fp: Path) -> None:
        super().__init__()
        self.fp = fp

    def exists(self) -> bool:
        if not self.fp.exists():
            logging.error(f"{self.fp} does not exist")
            return False
        if self.fp.stat().st_size == 0:
            logging.error(f"{self.fp} is empty")
            return False
        return True

    @abstractmethod
    def parse(self) -> None:
        pass

    @abstractmethod
    def write(self) -> None:
        pass


class FastaFile(File):
    def __init__(self, fp: Path) -> None:
        super().__init__(fp)

        try:
            stem = self.fp.stem.split(".")
            self.sample = stem[0]
            self.bin_name = stem[1]
        except IndexError:
            raise ValueError(
                f"FASTA filename {self.fp} not of format {{sample}}.{{bin_name}}.fasta/.fa/.fna"
            )

    def parse(self) -> Iterator[Tuple[str, str]]:
        with open(self.fp) as f:
            faiter = (x[1] for x in groupby(f, lambda line: line[0] == ">"))

            for header in faiter:
                # drop the ">"
                header_str = header.__next__()[1:].strip().split(" ")[0]

                # join all sequence lines to one.
                seq_str = "".join(s.strip() for s in faiter.__next__())

                yield (header_str, seq_str)

    def write(self):
        pass


class SamFile(File):
    def __init__(self, fp: Path) -> None:
        super().__init__(fp)
        stem = self.fp.stem.split("_")
        self.sample = stem[0]
        self.bin_name = stem[1].split(".")[0]

    def parse(self) -> Iterator[Dict[str, Union[str, int]]]:
        with open(self.fp, "r") as f:
            for line in f:
                if line.startswith("@"):
                    continue  # Skip header lines
                fields = line.split("\t")
                read_name = fields[0]
                flag = int(fields[1])
                reference_name = fields[2]
                position = int(fields[3])
                mapping_quality = int(fields[4])
                cigar = fields[5]
                mismatch = 0
                try:
                    if reference_name != "*":
                        # Find XM field for mismatches
                        for i in [13, 14, 15, 12, 11]:
                            if fields[i].split(":") == "XM":
                                mismatch = int(fields[i].split(":")[2])
                                break
                except IndexError:
                    pass

                parsed_read = {
                    "read_name": read_name,
                    "flag": flag,
                    "reference_name": reference_name,
                    "position": position,
                    "mapping_quality": mapping_quality,
                    "cigar": cigar,
                    "mismatch": mismatch,
                }
                yield parsed_read

    def parse_contig_lengths(self) -> Dict[str, int]:
        lengths = {}
        with open(self.fp, "r") as sam_file:
            for line in sam_file:
                if line.startswith("@"):
                    if line.startswith("@SQ"):
                        contig_name = line.split("\t")[1][3:]
                        contig_length = int(line.split("\t")[2][3:])
                        lengths[contig_name] = contig_length
                else:
                    break

        return lengths

    def write(self):
        pass


class Cov3File(File):
    def __init__(
        self,
        fp: Path,
        bin_name: str,
        mapq_cutoff: int = 5,
        mapl_cutoff: int = 50,
        max_mismatch_ratio: float = 0.03,
    ) -> None:
        super().__init__(fp)
        self.bin_name = bin_name

        self.mapq_cutoff = mapq_cutoff
        self.mapl_cutoff = mapl_cutoff
        self.max_mismatch_ratio = max_mismatch_ratio

        if not (0 <= self.mapq_cutoff <= 30):
            raise ValueError(
                f"MapQ cutoff of {self.mapq_cutoff} is not between 0 and 30"
            )
        if not (30 <= self.mapl_cutoff <= 80):
            raise ValueError(
                f"MapL cutoff of {self.mapl_cutoff} is not between 30 and 80"
            )
        if not (0.01 <= self.max_mismatch_ratio <= 0.3):
            raise ValueError(
                f"Max mismatch ratio of {self.max_mismatch_ratio} is not between 0.01 and 0.30"
            )

    def parse(self) -> Iterator[Dict[str, Union[str, int, float]]]:
        with open(self.fp) as f:
            for line in f.readlines():
                fields = line.split(",")
                yield {
                    "log_cov": float(fields[0]),
                    "GC_content": float(fields[1]),
                    "sample": fields[2],
                    "contig": fields[3],
                    "length": int(fields[4]),
                }

    def parse_sample_contig(self) -> Iterator[Dict[str, Union[str, int, List[float]]]]:
        with open(self.fp) as f:
            data_dict = {}
            for line in f.readlines():
                fields = line.split(",")
                parsed_line = {
                    "log_cov": float(fields[0]),
                    "GC_content": float(fields[1]),
                    "sample": fields[2],
                    "contig": fields[3],
                    "length": int(fields[4]),
                }

                sample = parsed_line["sample"]
                contig = parsed_line["contig"]

                if (sample, contig) not in data_dict:
                    data_dict[(sample, contig)] = {
                        "sample": sample,
                        "contig": contig,
                        "contig_length": parsed_line["length"],
                        "log_covs": [],
                        "GC_contents": [],
                    }

                data_dict[(sample, contig)]["log_covs"].append(parsed_line["log_cov"])
                data_dict[(sample, contig)]["GC_contents"].append(
                    parsed_line["GC_content"]
                )

            for values in data_dict.values():
                yield values

    def write(
        self, sams: List[SamFile], fasta: FastaFile, window_params: Dict[str, int]
    ) -> None:
        sam_generators = {sam.fp.stem: sam.parse() for sam in sams}
        cov3_generator = Cov3Generator(
            sam_generators,
            fasta.parse(),
            fasta.sample,
            fasta.bin_name,
            window_params,
            self.mapq_cutoff,
            self.mapl_cutoff,
            self.max_mismatch_ratio,
        )

        with open(self.fp, "w") as f_out:
            f_out.write("")  # Write header
            for line in cov3_generator.generate_cov3():
                f_out.write(",".join([str(v) for v in line.values()]))
                f_out.write("\n")



import os
from pathlib import Path
from src.pycov3.Directory import SamDir
from tests.unit.utils import create_sample_sam_file

dummy_sam_files = [
    "sample1_001.sam",
    "sample2_002.sam",
    "sample2_001.sam",
    "sample1_002.sam",
]


def test_sam_dir_init():
    # Define a temporary directory path for testing
    temp_dir = Path("sam_dir")
    os.makedirs(temp_dir)

    for dummy_file in dummy_sam_files:
        create_sample_sam_file(temp_dir / dummy_file)

    sam_dir = SamDir(temp_dir, False)

    assert len(sam_dir.files) == 4  # Assuming 4 dummy files were created


def test_sam_dir_get_bin():
    # Define a temporary directory path for testing
    temp_dir = Path("sam_dir")
    os.makedirs(temp_dir)

    for dummy_file in dummy_sam_files:
        create_sample_sam_file(temp_dir / dummy_file)

    sam_dir = SamDir(temp_dir, False)

    # Test the get_bin method
    sam_files = sam_dir.get_bin("001")

    assert len(sam_files) == 2
    assert sam_files[0].bin_name == "001"


def test_sam_dir_calculate_edge_length():
    # Define a temporary directory path for testing
    temp_dir = Path("sam_dir")
    os.makedirs(temp_dir)

    for dummy_file in dummy_sam_files:
        create_sample_sam_file(temp_dir / dummy_file)

    sam_dir = SamDir(temp_dir, False)

    window_edge_length = sam_dir.calculate_edge_length()

    assert window_edge_length == 30



from pathlib import Path
from src.pycov3.File import FastaFile
from tests.unit.utils import create_sample_fasta_file


def test_fasta_file_init():
    # Test FastaFile initialization
    fasta_file_path = Path("max_bin.001.fasta")
    create_sample_fasta_file(fasta_file_path)

    fasta_file = FastaFile(fasta_file_path)
    assert fasta_file.sample == "max_bin"
    assert fasta_file.bin_name == "001"


def test_fasta_file_parse():
    # Test FastaFile parse method
    fasta_file_path = Path("max_bin.001.fasta")
    create_sample_fasta_file(fasta_file_path)

    fasta_file = FastaFile(fasta_file_path)
    contigs = list(fasta_file.parse())

    assert len(contigs) == 2
    assert contigs[0][0] == "Contig1"
    assert len(contigs[0][1]) >= 50
    assert contigs[1][0] == "Contig2"


def test_fasta_file_write():
    # Test FastaFile write method (not implemented)
    pass



import pytest
from src.pycov3.Sequence import Window


def test_valid_initialization():
    # Test valid initialization of Window
    window = Window(0, 10, "ATGCATCGTA")
    assert window.start == 0
    assert window.end == 10
    assert window.len == 10


def test_invalid_start():
    # Test initialization with an invalid start
    with pytest.raises(ValueError):
        Window(-1, 10, "ATGCATCGTA")


def test_invalid_end():
    # Test initialization with an invalid end
    with pytest.raises(ValueError):
        Window(0, 0, "ATGCATCGTA")


def test_invalid_sequence_length():
    # Test initialization with an invalid sequence length
    with pytest.raises(ValueError):
        Window(0, 10, "ATGC")


def test_get_window():
    # Test getting the subsequence represented by the window
    window = Window(2, 6, "ATGC")
    sub_seq = window.get_window("ATGCATCGTA")
    assert sub_seq == "GCAT"


def test_generate_windows():
    # Test generating windows
    windows = Window.generate_windows(30, 5, 10, 2)
    assert len(windows) == 6


def test_generate_windows_invalid():
    # Test generating windows with invalid parameters
    assert not Window.generate_windows(20, 5, 15, 2)


def test_calculate_GC_content():
    # Test calculating GC content
    gc_content = Window.calculate_GC_content("ATGCATCGTA")
    assert gc_content == 0.4


def test_calculate_GC_skew():
    # Test calculating GC skew
    gc_skew = Window.calculate_GC_skew("ATGCATCGTA")
    assert gc_skew == [0, 0, 1, 0, 0, 0, -1, 0, 0, 0]



import logging
from pathlib import Path
from src.pycov3.Directory import SamDir, FastaDir, Cov3Dir


def test_sim_data():
    logging.basicConfig()
    logging.getLogger().setLevel(10)

    sim_fp = Path("sim")
    sams_fp = sim_fp / "sams"
    fastas_fp = sim_fp / "fastas"
    output_fp = sim_fp / "output"

    overwrite = True

    sam_d = SamDir(sams_fp, overwrite)

    window_params = {
        "window_size": 500,
        "window_step": 10,
        "edge_length": sam_d.calculate_edge_length(),
    }
    coverage_params = {
        "mapq_cutoff": None,
        "mapl_cutoff": None,
        "max_mismatch_ratio": None,
    }
    window_params = {k: v for k, v in window_params.items() if v is not None}
    coverage_params = {k: v for k, v in coverage_params.items() if v is not None}

    fasta_d = FastaDir(fastas_fp, overwrite)

    cov3_d = Cov3Dir(
        output_fp, overwrite, fasta_d.get_filenames(), window_params, coverage_params
    )

    cov3_d.generate(sam_d, fasta_d)

    # import shutil
    # shutil.copyfile(output_fp / "max_bin.002.cov3", "/mnt/d/Penn/pycov3/tests/data/sim/expected_output/max_bin.002.cov3")

    cov3_1 = cov3_d.get_bin("001")
    # cov3_2 = cov3_d.get_bin("002")

    sample_contigs_1 = list(cov3_1.parse_sample_contig())
    print(sample_contigs_1[0]["sample"])
    print(sample_contigs_1[0]["contig"])
    akk_001_k141_0_test(sample_contigs_1[0]["log_covs"])
    bfrag_001_k141_0_test(sample_contigs_1[1]["log_covs"])
    akk_001_k141_2_test(sample_contigs_1[2]["log_covs"])
    bfrag_001_k141_2_test(sample_contigs_1[3]["log_covs"])


def akk_001_k141_0_test(lc):
    assert sum(lc[: int(len(lc) / 2)]) < sum(lc[int(len(lc) / 2) :])


def bfrag_001_k141_0_test(lc):
    assert sum(lc[: int(len(lc) / 2)]) > sum(lc[int(len(lc) / 2) :])


def akk_001_k141_2_test(lc):
    assert (
        round(lc[0], 1) == round(lc[int(len(lc) / 2)], 1) == round(lc[len(lc) - 1], 1)
    )


def bfrag_001_k141_2_test(lc):
    assert sum(lc[: int(len(lc) / 2)]) < sum(lc[int(len(lc) / 2) :])



def create_sample_sam_file(sam_file_path):
    # Create a sample SAM file for testing
    sample_sam_content = (
        "@SQ\tSN:Contig1\tLN:100\n"
        "read1\t0\tContig1\t1\t30\t30M\t*\t0\t0\tATCGATCGATATCGATCGATATCGATCGAT\n"
        "read2\t0\tContig1\t11\t30\t30M\t*\t0\t0\tGCTAGCTAGCGCTAGCTAGCGCTAGCTAGC\n"
    )
    with open(sam_file_path, "w") as f:
        f.write(sample_sam_content)


def create_sample_fasta_file(fasta_file_path):
    # Create a sample fasta file for testing
    sample_fasta_content = ">Contig1\nATCGATCGATCGATCGATATCGATCGATATCGATCGATGCTAGCTAGCGCTAGCTAGCGCTAGCTAGC\n>Contig2\nGCTAGCTAGCTAGCTAGCGCTAGCTAGCGCTAGCTAGCATCGATCGATATCGATCGATATCGATCGAT\n"
    with open(fasta_file_path, "w") as f:
        f.write(sample_fasta_content)


def create_sample_cov3_file(cov3_file_path):
    # Create a sample cov3 file for testing
    sample_cov3_content = [
        "1.234,0.567,sample1,contig1,100\n",
        "1.234,0.567,sample1,contig1,100\n",
        "2.345,0.678,sample1,contig2,150\n",
        "3.456,0.789,sample2,contig2,150\n",
    ]
    with open(cov3_file_path, "w") as f:
        f.writelines(sample_cov3_content)



import pytest
from src.pycov3.Sequence import Contig


def test_valid_initialization():
    # Test valid initialization of Contig
    contig = Contig("Name", "ATGC", "Sample", "001", 10)
    assert contig.name == "Name"
    assert contig.seq_len == 4
    assert contig.sample == "Sample"
    assert contig.bin_name == "001"
    assert contig.edge_length == 10
    assert contig.window_size == 5000
    assert contig.window_step == 100


def test_invalid_window_step():
    # Test initialization with an invalid window step
    with pytest.raises(ValueError):
        Contig("Name", "ATGC", "Sample", "Bin", 10, window_step=5)


def test_invalid_window_size():
    # Test initialization with an invalid window size
    with pytest.raises(ValueError):
        Contig("Name", "ATGC", "Sample", "Bin", 10, window_size=100)


def test_invalid_window_size_step_relationship():
    # Test initialization with invalid window size and step relationship
    with pytest.raises(ValueError):
        Contig("Name", "ATGC", "Sample", "Bin", 10, window_size=200, window_step=150)


def test_invalid_window_size_divisibility():
    # Test initialization with window size not divisible by window step
    with pytest.raises(ValueError):
        Contig("Name", "ATGC", "Sample", "Bin", 10, window_size=1000, window_step=150)


def test_generate_windows():
    # Test generating windows
    contig = Contig(
        "Name", "ATGCGATCG" * 100, "Sample", "Bin", 5, window_size=500, window_step=100
    )
    assert len(contig.windows) == 4


def test_no_windows():
    # Test no windows when sequence length is insufficient
    contig = Contig("Name", "ATG", "Sample", "Bin", 5)
    assert len(contig.windows) == 0



from pathlib import Path
from src.pycov3.File import SamFile
from tests.unit.utils import create_sample_sam_file


def test_sam_file_init():
    # Test SamFile initialization
    sam_file_path = Path("Akk_001.sam")
    create_sample_sam_file(sam_file_path)

    sam_file = SamFile(sam_file_path)
    assert sam_file.sample == "Akk"
    assert sam_file.bin_name == "001"


def test_sam_file_parse():
    # Test SamFile parse method
    sam_file_path = Path("Akk_001.sam")
    create_sample_sam_file(sam_file_path)

    sam_file = SamFile(sam_file_path)
    parsed_reads = list(sam_file.parse())

    assert len(parsed_reads) == 2
    assert parsed_reads[0]["read_name"] == "read1"
    assert parsed_reads[0]["flag"] == 0
    assert parsed_reads[0]["reference_name"] == "Contig1"
    assert parsed_reads[0]["position"] == 1
    assert parsed_reads[0]["mapping_quality"] == 30
    assert parsed_reads[0]["cigar"] == "30M"
    assert parsed_reads[0]["mismatch"] == 0


def test_sam_file_parse_contig_lengths():
    # Test SamFile parse_contig_lengths method
    sam_file_path = Path("Akk_001.sam")
    create_sample_sam_file(sam_file_path)

    sam_file = SamFile(sam_file_path)
    contig_lengths = sam_file.parse_contig_lengths()

    assert len(contig_lengths) == 1
    assert contig_lengths["Contig1"] == 100


def test_sam_file_write():
    # Test SamFile write method (not implemented)
    pass



import logging


class Window:
    def __init__(self, start: int, end: int, sub_seq: str) -> None:
        self.start = start
        self.end = end
        self.len = self.end - self.start

        if self.start < 0:
            raise ValueError(f"Window start must be larger than 0 ({self.start} > 0)")
        if self.len <= 0:
            raise ValueError(
                f"Window end must be larger than start ({self.end > self.start})"
            )
        if self.len != len(sub_seq):
            raise ValueError(
                "Derived length from start and end not equal to length of given string"
            )

        self.gc_content = self.calculate_GC_content(sub_seq)
        # self.gc_skew = self.calculate_GC_skew(sub_seq)

    def get_window(self, seq: str):
        if self.end > len(seq):
            raise ValueError(
                f"Window end must be smaller than length of sequence ({self.end} < {len(seq)})"
            )
        return seq[self.start : self.end]

    @staticmethod
    def generate_windows(
        seq_len: int, edge_length: int, window_size: int, window_step: int
    ) -> list:
        return [
            (x, x + window_size)
            for x in range(
                edge_length, seq_len - edge_length - window_size + 1, window_step
            )
        ]

    @staticmethod
    def calculate_GC_content(seq: str) -> float:
        gc_count = seq.count("G") + seq.count("C")
        total_count = len(seq)
        gc_content = gc_count / total_count
        return gc_content

    @staticmethod
    def calculate_GC_skew(seq: str):
        gc_skew = [0]
        for i in range(1, len(seq)):
            gc_skew.append(gc_skew[i - 1] + (seq[i] == "G") - (seq[i] == "C"))
        return gc_skew


class Contig:
    def __init__(
        self,
        name: str,
        seq: str,
        sample: str,
        bin_name: str,
        edge_length: int,
        window_size: int = 5000,
        window_step: int = 100,
    ) -> None:
        self.name = name
        self.seq_len = len(seq)
        self.sample = sample
        self.bin_name = bin_name

        self.edge_length = edge_length
        self.window_size = window_size
        self.window_step = window_step

        if not (10 <= self.window_step <= 1000):
            raise ValueError(
                f"Window step of {self.window_step} is not between 10 and 1,000"
            )
        if not (500 <= self.window_size <= 10000):
            raise ValueError(
                f"Window size of {self.window_size} is not between 500 and 10,000"
            )
        if self.window_size < self.window_step * 2:
            raise ValueError("Window size must be at least twice the window step value")
        if self.window_size % self.window_step != 0:
            raise ValueError("Window step must evenly divide window size")

        if self.seq_len >= self.window_size + 2 * self.edge_length:
            self.windows = [
                Window(s, e, seq[s:e])
                for s, e in Window.generate_windows(
                    self.seq_len, self.edge_length, self.window_size, self.window_step
                )
            ]
            logging.debug(f"Num windows: {len(self.windows)}")
        else:
            self.windows = []



import os
from pathlib import Path
from src.pycov3.Directory import FastaDir
from tests.unit.utils import create_sample_fasta_file

dummy_fasta_files = [
    "max_bin.001.fasta",
    "max_bin.002.fa",
    "max_bin.003.fna",
    "max_bin.004.fasta",
]


def test_fasta_dir_init():
    # Define a temporary directory path for testing
    temp_dir = Path("fasta_dir")
    os.makedirs(temp_dir)

    for dummy_file in dummy_fasta_files:
        create_sample_fasta_file(temp_dir / dummy_file)

    # Initialize a FastaDir instance
    fasta_dir = FastaDir(temp_dir, False)

    # Test if the FastaDir object is created correctly
    assert len(fasta_dir.files) == 4  # Assuming 4 dummy files were created


def test_fasta_dir_get_bin():
    # Define a temporary directory path for testing
    temp_dir = Path("fasta_dir")
    os.makedirs(temp_dir)

    for dummy_file in dummy_fasta_files:
        create_sample_fasta_file(temp_dir / dummy_file)

    # Initialize a FastaDir instance
    fasta_dir = FastaDir(temp_dir, False)

    # Test the get_bin method
    fasta_file = fasta_dir.get_bin("004")

    assert fasta_file.sample == "max_bin"
    assert fasta_file.bin_name == "004"



import argparse
import logging
import sys
from pathlib import Path

from .Directory import FastaDir, SamDir, Cov3Dir

###
# Main steps:
# 1. Check and record parameters designated by user
# 2. Record GC content in sliding windows for contigs
# 	key subroutine:
# 	&GC_count
# 3. Distribute SAM files to threads and calculate coverages in sliding windows
# 	key subroutine:
# 	&sam_cov_parallel
# 4. Convert COV2 (coverage in samples) to COV3 (coverage in contig clusters)
# 5. Distribute COV3 files to threads and invoke R to estimate growth rates
# 	key subroutine:
# 	&cov3_estPTR_parallel
# 6. Output
###


def main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument("-S", "--sam_dir", help="the directory containing the sam file(s)")
    p.add_argument(
        "-F",
        "--fasta_dir",
        help="the directory containing the binned fasta file(s)",
    )
    p.add_argument(
        "-O",
        "--out_dir",
        help="the output directory",
    )
    p.add_argument(
        "-X",
        "--overwrite",
        help="overwrites any existing outputs",
        action="store_true",
    )

    p.add_argument(
        "-W",
        "--window_size",
        type=int,
        help="size (nt) of window for calculation of coverage (default: 5000)",
    )
    p.add_argument(
        "-D",
        "--window_step",
        type=int,
        help="step (nt) of window for calculation of coverage (default: 100)",
    )
    p.add_argument(
        "-M",
        "--mapq_cutoff",
        type=int,
        help="cutoff of mapping quality when calculating coverages (default: 5)",
    )
    p.add_argument(
        "-L",
        "--mapl_cutoff",
        type=int,
        help="cutoff of mapping length when calculating coverages (default: 50)",
    )
    p.add_argument(
        "-R",
        "--max_mismatch_ratio",
        type=float,
        help="maximum of mismatch ratio for each read as a hit (default: 0.03)",
        default=0.03,
    )
    p.add_argument(
        "-T",
        "--thread_num",
        type=int,
        help="set number of threads for parallel running (default: 1)",
        default=1,
    )

    p.add_argument(
        "-G",
        "--log_level",
        type=int,
        help="Sets the log level, default is info, 10 for debug (Default: 20)",
        default=20,
    )

    args = p.parse_args(argv)
    if not (args.sam_dir and args.fasta_dir and args.out_dir):
        p.print_help(sys.stderr)
        sys.exit(1)
    logging.basicConfig()
    logging.getLogger().setLevel(args.log_level)

    sam_d = SamDir(Path(args.sam_dir), args.overwrite)

    window_params = {
        "window_size": args.window_size,
        "window_step": args.window_step,
        "edge_length": sam_d.calculate_edge_length(),
    }
    coverage_params = {
        "mapq_cutoff": args.mapq_cutoff,
        "mapl_cutoff": args.mapl_cutoff,
        "max_mismatch_ratio": args.max_mismatch_ratio,
    }
    window_params = {k: v for k, v in window_params.items() if v is not None}
    coverage_params = {k: v for k, v in coverage_params.items() if v is not None}

    fasta_d = FastaDir(Path(args.fasta_dir), args.overwrite)

    cov3_d = Cov3Dir(
        Path(args.out_dir),
        args.overwrite,
        fasta_d.get_filenames(),
        window_params,
        coverage_params,
    )

    cov3_d.generate(sam_d, fasta_d)



import logging
from abc import ABC
from pathlib import Path
from typing import Dict, List, Union

from .File import FastaFile, SamFile, Cov3File


class Directory(ABC):
    def __init__(self, fp: Path, overwrite: bool) -> None:
        super().__init__()
        self.fp = fp.resolve()
        self.overwrite = overwrite
        self.files = []

        if not fp.is_dir() and fp.exists():
            raise ValueError(f"{self.fp} is not a directory")

    def get_filenames(self) -> List[str]:
        return [".".join(f.fp.name.split(".")[:-1]) for f in self.files]


class FastaDir(Directory):
    def __init__(self, fp: Path, overwrite: bool) -> None:
        super().__init__(fp, overwrite)

        self.files = [
            FastaFile(x)
            for x in self.fp.iterdir()
            if str(x).endswith((".fasta", ".fa", ".fna"))
        ]
        if not self.files:
            raise ValueError(
                f"No files found ending in .fasta, .fa, or .fna in {self.fp}"
            )

    def get_bin(self, bin_name: str) -> FastaFile:
        fasta_l = [f for f in self.files if f.bin_name == bin_name]
        if len(fasta_l) != 1:
            raise ValueError(
                f"Found 0 or more than 1 matches for bin {bin_name} in FASTAs"
            )
        return fasta_l[0]


class SamDir(Directory):
    def __init__(self, fp: Path, overwrite: bool) -> None:
        super().__init__(fp, overwrite)

        self.files = [SamFile(x) for x in self.fp.iterdir() if str(x).endswith(".sam")]
        if not self.files:
            raise ValueError(f"No files found ending in .sam in {self.fp}")

    def calculate_edge_length(self) -> int:
        edge_length = 0
        for sam in self.files:
            checks = 0
            for line in sam.parse():
                cutoff = 0
                for index, char in enumerate(line["cigar"]):
                    if not char.isdigit():
                        cutoff = index
                        break

                try:
                    cigar_val = int(line["cigar"][:cutoff])
                    if cigar_val > edge_length:
                        edge_length = cigar_val
                except ValueError:
                    logging.debug(f"Couldn't parse CIGAR value of {line['cigar']}")
                    break

                checks += 1
                if checks > 100:
                    break

        return edge_length

    def get_bin(self, bin_name: str) -> List[SamFile]:
        return [s for s in self.files if s.bin_name == bin_name]


class Cov3Dir(Directory):
    def __init__(
        self,
        fp: Path,
        overwrite: bool,
        fns: List[str],
        window_params: Dict[str, int],
        coverage_params: Dict[str, Union[int, float]],
    ) -> None:
        super().__init__(fp, overwrite)
        self.window_params = window_params

        if not fp.exists():
            logging.info(f"{self.fp} does not exist, creating it now")
            self.fp.mkdir(parents=True, exist_ok=True)
        if any(self.fp.iterdir()) and not self.overwrite:
            raise ValueError(
                f"{self.fp} is a non-empty directory, please either point output to an empty or non-existent directory or run with the overwrite flag"
            )

        self.files = [
            Cov3File(fp / f"{fn}.cov3", fn.split(".")[-1], **coverage_params)
            for fn in fns
        ]

    def generate(self, sam_d: SamDir, fasta_d: FastaDir) -> None:
        for cov3 in self.files:
            cov3.write(
                sam_d.get_bin(cov3.bin_name),
                fasta_d.get_bin(cov3.bin_name),
                self.window_params,
            )

    def get_bin(self, bin_name: str) -> Cov3File:
        result = [c for c in self.files if c.bin_name == bin_name]
        assert len(result) == 1, "There should be exactly one cov3 file for each bin"
        return result[0]



import logging
import math
from collections import OrderedDict
from typing import Dict, Iterator, List, Tuple, Union

from .Sequence import Contig


class Cov3Generator:
    def __init__(
        self,
        sam_generators: Dict[str, Iterator[Dict[str, Union[str, int]]]],
        fasta_generator: Iterator[Tuple[str, str]],
        sample: str,
        bin_name: str,
        window_params: Dict[str, int],
        mapq_cutoff: int,
        mapl_cutoff: int,
        max_mismatch_ratio: float,
    ) -> None:
        self.sam_generators = sam_generators
        self.fasta_generator = fasta_generator
        self.sample = sample
        self.bin_name = bin_name
        self.window_params = window_params
        self.mapq_cutoff = mapq_cutoff
        self.mapl_cutoff = mapl_cutoff
        self.max_mismatch_ratio = max_mismatch_ratio

        self.min_cov_window = 0.1
        self.min_window_count = 5

    def generate_cov3(self) -> Iterator[Dict[str, Union[str, int, float]]]:
        next_lines = OrderedDict(
            sorted(
                {name: next(sg, {}) for name, sg in self.sam_generators.items()}.items()
            )
        )

        for contig_name, seq in self.fasta_generator:
            contig = Contig(
                contig_name, seq, self.sample, self.bin_name, **self.window_params
            )
            logging.debug(f"Current contig: {contig_name}")

            for name, line in next_lines.items():
                mut_line = line
                coverages = {}
                while True:
                    if not mut_line:
                        break  # Generator is exhausted
                    if mut_line["reference_name"] == "*":
                        next_lines[name] = {}
                        break  # SAM file has no more mapped reads
                    if mut_line["reference_name"] != contig_name:
                        break  # This contig is unmapped by this SAM file
                    if contig.windows:
                        coverages = self.__update_coverages(
                            coverages,
                            mut_line,
                            contig.edge_length,
                            contig.window_step,
                        )
                    mut_line = next(self.sam_generators[name], {})

                next_lines[
                    name
                ] = mut_line  # Instead of updating with every iteration of the while loop
                if coverages:
                    for info in self.__log_cov_info(
                        contig,
                        coverages,
                        contig.edge_length,
                        contig.window_size,
                        contig.window_step,
                    ):
                        info["sample"] = name
                        info["contig"] = contig_name
                        info["length"] = contig.seq_len
                        yield info

    def __update_coverages(
        self,
        coverages: Dict[int, int],
        line: Dict[str, Union[str, int]],
        edge_length: int,
        window_step: int,
    ) -> Dict[int, int]:
        mapl = self.calculate_mapl(line["cigar"])
        if (
            line["mapping_quality"] >= self.mapq_cutoff
            and mapl >= self.mapl_cutoff
            and line["mismatch"] <= self.max_mismatch_ratio * mapl
        ):
            start_step = int((line["position"] - 1 - edge_length) / window_step)
            end_step = int((line["position"] - 1 + mapl - edge_length) / window_step)

            if start_step not in coverages:
                coverages[start_step] = 0
            coverages[start_step] += window_step - (
                (line["position"] - 1 - edge_length) % window_step
            )
            if end_step not in coverages:
                coverages[end_step] = 0
            coverages[end_step] += (
                line["position"] - 1 + mapl - edge_length
            ) % window_step

            for step in range(start_step + 1, end_step):
                if step not in coverages:
                    coverages[step] = 0
                coverages[step] += window_step

        return coverages

    def __log_cov_info(
        self,
        contig: Contig,
        coverages: Dict[int, int],
        edge_length: int,
        window_size: int,
        window_step: int,
    ) -> List[Dict[str, float]]:
        first_i = contig.windows[0].start
        last_i = contig.windows[-1].end
        first_step = int((first_i - 1 - edge_length) / window_step)
        last_step = int((last_i - 1 - edge_length) / window_step)

        cov_step = []
        cov_window_sum = 0
        qualified_info = []  # Information to output
        n = 0  # Window index

        for step in range(first_step, last_step):
            if step in coverages.keys():
                cov_step.append(coverages[step])
                cov_window_sum += coverages[step]
            else:
                cov_step.append(0)

            if len(cov_step) == window_size / window_step:
                avg_cov_window = cov_window_sum / window_size
                window = contig.windows[n]
                gc_content = window.gc_content
                n += 1
                cov_window_sum -= cov_step.pop(0)
                if avg_cov_window > self.min_cov_window:
                    log_cov = round(math.log(avg_cov_window) / math.log(2), 4)
                    qualified_info.append(
                        {"log_cov": log_cov, "GC_content": gc_content}
                    )

        if n >= self.min_window_count and len(qualified_info) == n:
            return qualified_info
        return []

    @staticmethod
    def calculate_mapl(cigar: str) -> int:
        operations = []
        current_length = ""

        if cigar == "*":
            return -1

        for char in cigar:
            if char.isdigit():
                current_length += char
            else:
                operations.append((int(current_length), char))
                current_length = ""

        return sum([n for n, c in operations if c == "M" or c == "D"]) - sum(
            [n for n, c in operations if c == "I"]
        )



from pathlib import Path
from src.pycov3.File import Cov3File
from tests.unit.utils import create_sample_cov3_file


def test_cov3_file_init():
    # Test cov3 file initialization
    cov3_file = Cov3File(Path("max_bin.001.cov3"), "001")
    assert cov3_file.bin_name == "001"


def test_cov3_file_parse():
    # Test cov3 file parse
    cov3_fp = Path("max_bin.001.cov3")
    create_sample_cov3_file(cov3_fp)
    cov3_file = Cov3File(cov3_fp, "001")

    cov3_vals = list(cov3_file.parse())

    assert len(cov3_vals) == 4
    assert cov3_vals[0]["log_cov"] == 1.234
    assert cov3_vals[0]["GC_content"] == 0.567
    assert cov3_vals[0]["sample"] == "sample1"
    assert cov3_vals[0]["contig"] == "contig1"
    assert cov3_vals[0]["length"] == 100


def test_cov3_file_parse_sample_contig():
    # Test cov3 file parse by sample and contig
    cov3_fp = Path("max_bin.001.cov3")
    create_sample_cov3_file(cov3_fp)
    cov3_file = Cov3File(cov3_fp, "001")

    cov3_vals = list(cov3_file.parse_sample_contig())

    assert len(cov3_vals) == 3
    assert cov3_vals[0]["sample"] == "sample1"
    assert cov3_vals[0]["contig"] == "contig1"
    assert len(cov3_vals[0]["log_covs"]) == 2
